
#include <iostream>
#include <mutex>
#include "Instantiation_and_registry.h"
#include "ASIO_device.h"


class Aggregate : public AudioHost {
public:
    ASIODevice dev1;
    ASIODevice dev2;

    int32_t* shared_buffer[2]{nullptr};
    int32_t* temp_buffer[2]{nullptr};
    std::mutex exchange_lock;
    bool exchange_ready = false;

    Aggregate(int device_index_1, int device_index_2) {
        dev1.attach(device_index_1, this);
        dev2.attach(device_index_2, this);

        if (dev1.get_buffer_size() != dev2.get_buffer_size() || dev1.get_sample_rate() != dev2.get_sample_rate()) {
            dev1.detach();
            dev2.detach();
            return;
        }

        int buffer_size = dev1.get_buffer_size();

        temp_buffer[0]   = new int32_t[buffer_size] { 0 };
        temp_buffer[1]   = new int32_t[buffer_size] { 0 };
        shared_buffer[0] = new int32_t[buffer_size] { 0 };
        shared_buffer[1] = new int32_t[buffer_size] { 0 };

        exchange_ready = true;
    }

    const bool& success() const {
        return exchange_ready;
    }
    
    void bufferSwitch(long dbi, bool directProcess, ASIODevice* caller) override {
        const int* mic    = caller-> get_input(0);
              int* main_L = caller->get_output(0);
              int* main_R = caller->get_output(1);
        const long buffersize = caller->get_buffer_size();
        const long buffersize_b = buffersize * sizeof(int32_t);

        // Perform internal transfer of the audio signal.
        memcpy(main_L, mic, buffersize_b);
        memcpy(main_R, mic, buffersize_b);

        // Do not try to perform the exchange if both devices are not ready.
        if (!exchange_ready) {
            caller->output_ready();
            return;
        }
        
        // Find which device we are working with
        int device = caller->get_device_index() == dev1.get_device_index();
        int other_device = !device;

        // Perform exchange!
        exchange_lock.lock();
        memcpy(shared_buffer[device], mic, buffersize_b); // Send data to the other device
        memcpy(temp_buffer[device], shared_buffer[other_device], buffersize_b); // Receive data from the other device
        exchange_lock.unlock();
        // Exchange complete

        // Now sum in the signal from the other device.
        memsum(main_L, temp_buffer[device], buffersize);
        memsum(main_R, temp_buffer[device], buffersize);

        caller->output_ready(); // Tell the current device that the output data is ready.
    }

    void shutdown() {
        dev1.detach();
        dev2.detach();
        delete(shared_buffer[0]);
        delete(shared_buffer[1]);
    }

private:

    void memsum(int* _Dst, const int* _Src, size_t _Size) {
        for (; _Size > 0; _Size--)
            *_Dst++ += *_Src++;
    }

};

bool choose_device_dialogue(int& device_index_1, int& device_index_2) {
    const int num_devices = ASIO_SDK::get_num_ASIO_drivers();
    std::wstring* names = new std::wstring[num_devices];
    ASIO_SDK::read_driver_name_all(names);

    int index = 0;
    std::wstring* name = &names[0];
    while (index++ - num_devices)
        std::wcout << index << ":" << *name++ << std::endl;
    
    std::cout << "Choose the first audio interface" << std::endl;
	std::string s;
	std::cin >> s;
	device_index_1 = strtol(s.c_str(), nullptr, 0) - 1;
	if (device_index_1 < 0 || device_index_1 >= num_devices) {
		std::cout << "Did not understand the input. Exiting..." << std::endl;
		return false;
	}

    std::cout << "Choose the second audio interface" << std::endl;
	std::cin >> s;
	device_index_2 = strtol(s.c_str(), nullptr, 0) - 1;
	if (device_index_2 < 0 || device_index_2 >= num_devices) {
		std::cout << "Did not understand the input. Exiting..." << std::endl;
		return false;
	}

    return true;
}

int main() {
    int index1 = -1;
    int index2 = -1;
    if (!choose_device_dialogue(index1, index2))
        return 1;
    
    Aggregate host(index1, index2);
    if (!host.success()) {
        std::cout << "Cannot run test. The sample rate and buffer size of each device must be the same.";
        return -1;
    }

    std::cout << "Test now running. Enter anything to stop the test" << std::endl;
	std::string s;
    std::cin >> s;
    std::cout << "Shutting down..." << std::endl;

    host.shutdown();

    return 0;
}