
#include <iostream>
#include <mutex>
#include "Instantiation_and_registry.h"
#include "ASIO_device.h"
#include "ExchangeBuffer.h"

#define PI 3.14159265

const double sinewave_hz = 220;
const int top = 0x7FFFFFFF;
void sineWave(int hz, float amplitude, int* buffer, int buffer_size, int sampleRate, long samplePos) {
	for (int i = 0; i < buffer_size; i++) {
		float value = amplitude * std::sin(PI * 2 * samplePos / sampleRate * hz);
		buffer[i] = (int)value;
		samplePos++;
	}
}


class Aggregate : public AudioHost {
public:
    ASIODevice dev1;
    ASIODevice dev2;
    ExchangeBuffer* transfer_1_2;
    ExchangeBuffer* transfer_2_1;

    int32_t* temp_buffer[2]{nullptr};
    std::mutex exchange_lock;
    bool exchange_ready = false;

    int call_difference = 0;
    int call_difference_last = 0;
	std::chrono::system_clock::time_point last_pair_change;
    bool first_pair_change = true;

    int max = 0;
    int min = 0;

    Aggregate(int device_index_1, int device_index_2) {
        dev1.attach(device_index_1, this);
        dev1.setChannelStatus_noCommit(false, 4, true);
        dev1.setChannelStatus_noCommit(false, 5, true);
        dev2.attach(device_index_2, this);

        int buffer_size = dev1.get_buffer_size();
        //transfer_1_2 = new ExchangeBuffer(buffer_size);
        transfer_2_1 = new ExchangeBuffer(buffer_size);
        temp_buffer[0] = new int32_t[buffer_size] {0};
        temp_buffer[1] = new int32_t[buffer_size] {0};

        exchange_ready = true;
    }

    void stop() {
        dev1.stop();
        dev2.stop();
    }

    void start() {
        dev1.start();
        dev2.start();
    }
    
    void bufferSwitch(long dbi, bool directProcess, ASIODevice* caller) override {
        if (caller->get_buffer_size() != 64 || caller->get_sample_rate() != 44100)
            throw std::exception("For this test to run, sample rate must be 44100Hz, and buffer size must be 64");

        const int32_t* mic    = caller-> get_input(0);
              int32_t* main_L = caller->get_output(0);
              int32_t* main_R = caller->get_output(1);
              int32_t* out_5  = nullptr;
              int32_t* out_6  = nullptr;
        const long buffersize = caller->get_buffer_size();
        const long buffersize_b = buffersize * sizeof(int32_t);

        if (caller->get_device_index() == dev1.get_device_index()) {
            out_5 = caller->get_output(4);
            out_6 = caller->get_output(5);
        }


        // Perform common processing for both devices
        memcpy(main_L, mic, buffersize_b);
        memcpy(main_R, mic, buffersize_b);
        if (out_5) memcpy(out_5,  mic, buffersize_b);
        if (out_6) memcpy(out_6,  mic, buffersize_b);

        if (!exchange_ready) {
            caller->output_ready();
            return;
        }
        
        // Find which device we are working with
        int device = caller->get_device_index() == dev1.get_device_index();
        int other_device = !device;

        const int call_diff_change = device ? 1 : -1;

        // Perform exchange!
        exchange_lock.lock();
        call_difference += call_diff_change;
        const int call_diff = call_difference;
        exchange_lock.unlock();

        if (device) {
            //transfer_1_2->writeNext(mic);
            transfer_2_1->readNext(temp_buffer[device]);
        } else {
            transfer_2_1->writeNext(mic);
            //transfer_1_2->readNext(temp_buffer[device]);
        }
        // Exchange complete

        // Now sum in the signal from the other device.
        memsum(main_L, temp_buffer[device], buffersize);
        memsum(main_R, temp_buffer[device], buffersize);
        if (out_5) memsum(out_5,  temp_buffer[device], buffersize);
        if (out_6) memsum(out_6,  temp_buffer[device], buffersize);
        // And send the data!
        caller->output_ready();


        // From this point onward, only device 1 will perform processing
        if (!device)
            return;
        

        if (call_diff != call_difference_last) {
            //std::cout << "Call difference: " << call_diff << std::endl;
            //if (abs(call_diff - call_difference_last) > 1)
                //std::cout << "Call difference jumped by more than 1 call! It jumped " << call_diff - call_difference_last << " calls" << endl;
            call_difference_last = call_diff;
        }

        if (max < call_diff) {
            max = call_diff;
            
            std::chrono::system_clock::time_point now = std::chrono::system_clock::now();
            if (!first_pair_change) {
                int diff = abs(std::chrono::duration_cast<std::chrono::microseconds>(now - last_pair_change).count());
                ///std::cout << "Buffer pair change interval: " << ((float) diff / 1000.0f) << "ms" << std::endl;
            }
            last_pair_change = now;
            first_pair_change = false;
            //*/
        }
        // min = call_diff < min ? call_diff : min;
        // max = call_diff > max ? call_diff : max;
        
    }

    void shutdown() {
        dev1.detach();
        dev2.detach();
        delete(temp_buffer[0]);
        delete(temp_buffer[1]);
        delete(transfer_1_2);
        delete(transfer_2_1);
    }
private:
    void memsum(int32_t* _Dst, const int32_t* _Src, size_t _Size) {
        for (; _Size > 0; _Size--)
            *_Dst++ += *_Src++;
    }
};

bool choose_device_dialogue(int& device_index_1, int& device_index_2) {
    const int num_devices = ASIO_SDK::get_num_ASIO_drivers();
    std::wstring* names = new std::wstring[num_devices];
    ASIO_SDK::read_driver_name_all(names);

    int index = 0;
    std::wstring* name = &names[0];
    while (index++ - num_devices)
        std::wcout << index << ":" << *name++ << std::endl;
    
    std::cout << "Choose the first audio interface" << std::endl;
	std::string s;
	std::cin >> s;
	device_index_1 = strtol(s.c_str(), nullptr, 0) - 1;
	if (device_index_1 < 0 || device_index_1 >= num_devices) {
		std::cout << "Did not understand the input. Exiting..." << std::endl;
		return false;
	}

    std::cout << "Choose the second audio interface" << std::endl;
	std::cin >> s;
	device_index_2 = strtol(s.c_str(), nullptr, 0) - 1;
	if (device_index_2 < 0 || device_index_2 >= num_devices) {
		std::cout << "Did not understand the input. Exiting..." << std::endl;
		return false;
	}

    return true;
}

int main() {
    int index1=9, index2=5;
    //if (!choose_device_dialogue(index1, index2))
        //return 1;
    
    Aggregate host(index1, index2);
    //host.stop();
    //host.bufferSwitch(0, true, &host.dev1);

    std::cout << "Test now running. Enter anything to stop the test" << std::endl;
	std::string s;
    std::cin >> s;
    std::cout << "Shutting down..." << std::endl;

    host.shutdown();

    return 0;
}