
#include <iostream>
#include <string>
#include <vector>
#include <thread>
#include <math.h>
#include <mutex>

#include "Instantiation_and_registry.h"
#include "ASIO_device.h"
#include "asio_sdk/ginclude.h"

#ifndef PI
#define PI 3.14159265
#endif

const int32_t top = 0x7FFFFFFF;
void sineWave(const int& hz, const float& amplitude, int* buffer, const int& buffer_size, const int& sampleRate, long samplePos) {
	for (int i = 0; i < buffer_size; i++) {
		float value = amplitude * std::sin(PI * 2 * samplePos / sampleRate * hz);
		buffer[i] = (int)value;
		samplePos++;
	}
}

class AudioEngine : public AudioHost {
private:
	long samplePos = 0L;
    long measuredBufferSize = 0;
    std::vector<int> measures;

    std::mutex lock;

public:
    void readAllMeasurements(std::vector<int>& target) {
        lock.lock();
        target.assign(measures.begin(), measures.end());
        measures.clear();
        lock.unlock();
    };

	void bufferSwitch(long doubleBufferIndex, bool directProcess, ASIODevice* caller) override {
        measuredBufferSize += caller->get_buffer_size();
        if (directProcess) {
            lock.lock();
            measures.push_back(measuredBufferSize);
            lock.unlock();
            measuredBufferSize = 0;
        }

        if (!directProcess)
            std::cout << "Skip!" << std::endl;

		int32_t* bufferL = caller->get_output(0);
		int32_t* bufferR = caller->get_output(1);

		sineWave(220, top * 0.125, bufferL, caller->get_buffer_size(), caller->get_sample_rate(), samplePos);
		memcpy(bufferR, bufferL, caller->get_buffer_size() * sizeof(int32_t));

		samplePos += caller->get_buffer_size();
	}

	long asioMessage(long selector, long value, void* message, double* opt, ASIODevice* caller) override {
		//cout << "#Selector1: " << selector << "   #Value: " << value << "   #Message: " << endl;

		char* message_char = (char*)message;

		// Device asking for supported selectors
		if (selector == kAsioSelectorSupported)
			switch (value) {
			case kAsioEngineVersion:	return 1L;
			case kAsioSupportsTimeInfo: return 1L;
			case kAsioResetRequest:		return 1L;
			case kAsioBufferSizeChange: return 1L;
			case kAsioResyncRequest:	return 0L;
			case kAsioLatenciesChanged: return 0L;
			case kAsioOverload:			return 0L;

			default:					return 0;
			}

		switch (selector) {
		case kAsioEngineVersion:	return 2L;
		case kAsioResetRequest:		return 0L;
		case kAsioResyncRequest:	return 0L;
		case kAsioSupportsTimeInfo: return 0L;
		case kAsioOverload:			return 0L;
		}

		return 0L;
	}

	ASIOTime* bufferSwitchTimeInfo(ASIOTime* params, long doubleBufferIndex, bool directProcess, ASIODevice* caller) override {
		return nullptr;
	}
};

int findIndexOfData(int data, std::vector<int>& buffer) {
    std::vector<int>::iterator it = buffer.begin();

    for (int i = 0; it != buffer.end(); it++, i++)
        if (*it == data)
            return i;
    
    return -1;
}

void print_process(bool* running, AudioEngine* engine) {
    *running = true;
    std::vector<int> data;
    std::vector<int> bufferSizes;
    std::vector<int> numBufferSizes;

    while (*running) {
        Sleep(500);
        engine->readAllMeasurements(data);

        int minimum = 0;
        for (int i = 0; i < data.size(); i++) {
            int index = findIndexOfData(data[i], bufferSizes);

            if (index == -1) {
                bufferSizes.push_back(data[i]);
                numBufferSizes.push_back(1);
            }
            else
                numBufferSizes[index] += 1;
        }

        for (int i = 0; i < bufferSizes.size(); i++)
            std::cout << std::to_string(bufferSizes[i]) << "spl=" << std::to_string(numBufferSizes[i]) << " | ";
        std::cout << std::endl;

        data.clear();
        bufferSizes.clear();
        numBufferSizes.clear();
    }
}

int ask_for_device(std::string query) {
    const int num_devices = ASIO_SDK::get_num_ASIO_drivers();
    std::wstring* names = new std::wstring[num_devices];
    ASIO_SDK::read_driver_name_all(names);

    int index = 0;
    std::wstring* name = &names[0];
    while (index++ - num_devices)
        std::wcout << index << ": " << *name++ << std::endl;
    std::cout << index << ": [EXIT]" << std::endl;
	std::cout << query << std::endl;

	std::string s;
	std::cin >> s;
	long ans = strtol(s.c_str(), nullptr, 0);
	if (ans <= 0 || ans > num_devices + 1) {
		std::cout << "Did not understand the input." << std::endl;
		return -1;
	}

	if (ans == index)
		return -2;

    return (int) ans - 1;
}

int main() {
    int device_index = -1; 
	while (device_index == -1)
		device_index = ask_for_device("Which device would you like to test?");
	
	if (device_index == -2)
		return 0;
	
	AudioEngine engine;
    ASIODevice device(device_index, &engine);

    bool running = true;
    std::thread print_thread(print_process, &running, &engine);

    std::cout << "Enter anything to end the test.";
    std::string s;
    std::cin >> s;

    running = false;
    device.detach();
    print_thread.join();
    
	return 0;
}