
#include <iostream>
#include <string>
#include <vector>
#include <math.h>
#include <mutex>

#include "Instantiation_and_registry.h"
#include "ASIO_device.h"
#include "asio_sdk/ginclude.h"

#ifndef PI
#define PI 3.14159265
#endif

const int32_t top = 0x7FFFFFFF;
void sineWave(const int& hz, const float& amplitude, int* buffer, const int& buffer_size, const int& sampleRate, long samplePos) {
	for (int i = 0; i < buffer_size; i++) {
		float value = amplitude * std::sin(PI * 2 * samplePos / sampleRate * hz);
		buffer[i] = (int)value;
		samplePos++;
	}
}

class AudioEngine : public AudioHost {
private:
	long samplePos = 0L;

	bool playing = false;
	std::mutex lock;

public:
	void toggle() {
		if (playing)
			stop();
		else
			play();
	}

	void play() {
		lock.lock();
		playing = true;
		lock.unlock();
	}

	void stop() {
		lock.lock();
		playing = false;
		lock.unlock();
	}

	void bufferSwitch(long doubleBufferIndex, bool directProcess, ASIODevice* caller) override {
		lock.lock();
		const bool run = playing;
		lock.unlock();

		if (!run) {
			caller->silence_output();
			caller->output_ready();
			samplePos = 0;
			return;
		}
		

		int32_t* bufferL = caller->get_output(0);
		int32_t* bufferR = caller->get_output(1);

		sineWave(220, top * 0.125, bufferL, caller->get_buffer_size(), caller->get_sample_rate(), samplePos);
		memcpy(bufferR, bufferL, caller->get_buffer_size() * sizeof(int32_t));

		samplePos += caller->get_buffer_size();
	}

	long asioMessage(long selector, long value, void* message, double* opt, ASIODevice* caller) override {
		//cout << "#Selector1: " << selector << "   #Value: " << value << "   #Message: " << endl;

		char* message_char = (char*)message;

		// Device asking for supported selectors
		if (selector == kAsioSelectorSupported)
			switch (value) {
			case kAsioEngineVersion:	return 1L;
			case kAsioSupportsTimeInfo: return 1L;
			case kAsioResetRequest:		return 1L;
			case kAsioBufferSizeChange: return 1L;
			case kAsioResyncRequest:	return 0L;
			case kAsioLatenciesChanged: return 0L;
			case kAsioOverload:			return 0L;

			default:					return 0;
			}

		switch (selector) {
		case kAsioEngineVersion:	return 2L;
		case kAsioResetRequest:		return 0L;
		case kAsioResyncRequest:	return 0L;
		case kAsioSupportsTimeInfo: return 0L;
		case kAsioOverload:			return 0L;
		}

		return 0L;
	}

	ASIOTime* bufferSwitchTimeInfo(ASIOTime* params, long doubleBufferIndex, bool directProcess, ASIODevice* caller) override {
		return nullptr;
	}
};

int ask_for_device() {
    const int num_devices = ASIO_SDK::get_num_ASIO_drivers();
    std::wstring* names = new std::wstring[num_devices];
    ASIO_SDK::read_driver_name_all(names);

    std::cout << "Choose a device:" << std::endl;
    int index = 0;
    std::wstring* name = &names[0];
    while (index++ - num_devices)
        std::wcout << index << ":" << *name++ << std::endl;
    
	std::string s;
	std::cin >> s;
	long ans = strtol(s.c_str(), nullptr, 0);
	if (ans <= 0 || ans > num_devices) {
		std::cout << "Did not understand the input.";
		return -1;
	}

    return (int) ans - 1;
}

int main() {
    int device_index_1 = ask_for_device();
    int device_index_2 = ask_for_device();
	
	AudioEngine engine_1;
	AudioEngine engine_2;

    ASIODevice device_L(device_index_1, &engine_1);
    ASIODevice device_R(device_index_2, &engine_2);

	int ans = 0;

	while (ans != -1) {
		std::cout << " 1: Toggle first  device" << std::endl;
		std::cout << " 2: Toggle second device" << std::endl;
		std::cout << "-1: Exit"                 << std::endl;

		std::string s;
		std::cin >> s;
		ans = strtol(s.c_str(), nullptr, 0);

		switch (ans) {
			case -1: break;
			case  1: engine_1.toggle(); break;
			case  2: engine_2.toggle(); break;
			default:
				std::cout << "Did not understand input: " << ans << std::endl;
		}
	}

	device_L.detach();
	device_R.detach();
	return 0;
}