



//   "-I\"${workspaceFolder}\\include\\parallel_asio\""

#include <iostream>
#include <string>
#include <math.h>
#include <chrono>
#include <mutex>
#include <format>

#include "Instantiation_and_registry.h"
#include "ASIO_device.h"
#include "asio_sdk/ginclude.h"

#define PI 3.14159265

using namespace std;

const double sinewave_hz = 220;
const int top = 0x7FFFFFFF;
void sineWave(int hz, float amplitude, int* buffer, int buffer_size, int sampleRate, long samplePos) {
	for (int i = 0; i < buffer_size; i++) {
		float value = amplitude * sin(PI * 2 * samplePos / sampleRate * hz);
		buffer[i] = (int)value;
		samplePos++;
	}
}

class DAW1 : public AudioHost {

public:
	int min = -1;
	int mean = -1;
	int max = -1;
	chrono::nanoseconds last_delay = chrono::nanoseconds(0);
	mutex lock;


	void getValues(int& min, int& mean, int& max) {
		lock.lock();
		min  = this->min;
		mean = this->mean;
		max  = this->max;

		this->min = -1;
		this->mean = -1;
		this->max = -1;
		lock.unlock();
	}

private:
	boolean first = true;
	int ignition_runs = 5;
    long samplePos = 0L;
	chrono::system_clock::time_point now = chrono::system_clock::now();
	chrono::system_clock::time_point last_now;

	void bufferSwitch(long doubleBufferIndex, bool directProcess, ASIODevice* caller) override {
		now = chrono::system_clock::now();

        if (ignition_runs) {
			ignition_runs--;
			caller->silence_output();
			caller->output_ready();
			return;
		}
		
		if (first) {
			//SetPriorityClass(GetCurrentProcess(), REALTIME_PRIORITY_CLASS);
			//cout << "First" << endl;
			last_now = now;
			first = false;
			caller->silence_output();
			caller->output_ready();
			return;
		}

		chrono::nanoseconds delay = chrono::duration_cast<chrono::nanoseconds>(now - last_now);

		if (last_delay == 0ns) {
			last_delay = delay;
			last_now = now;
			caller->output_ready();
			return;
		}

		int diff = abs(chrono::duration_cast<chrono::nanoseconds>(delay - last_delay).count());

		lock.lock();
		if (mean == -1)
			mean = diff;
		else
			mean = (mean + diff) / 2;
		if (max < diff || max == -1)
			max = diff;
		if (min > diff || min == -1)
			min = diff;
		lock.unlock();

		last_delay = delay;
		last_now = now;


		sineWave(sinewave_hz, top * 0.125, caller->get_output(0), caller->get_buffer_size(), caller->get_sample_rate(), samplePos);
		memcpy(caller->get_output(1), caller->get_output(0), sizeof(int32_t) * caller->get_buffer_size());
        samplePos += caller->get_buffer_size();
	}
};

void print_process(bool* running, DAW1* engine) {
	*running = true;
	int min=0, max=0, mean=0;
	float min_f=0.0f, max_f=0.0f, mean_f=0.0f;
	float div = 1000.0f;

	while(*running) {
		Sleep(500);
		engine->getValues(min, mean, max);

		min_f = (float) min / div;
		mean_f = (float) mean / div;
		max_f = (float) max / div;

		printf("%10.1f | %10.1f | %10.1f", min_f, mean_f, max_f);
		cout << endl;
	}
}

int ask_for_device(string query) {
    const int num_devices = ASIO_SDK::get_num_ASIO_drivers();
    wstring* names = new wstring[num_devices];
    ASIO_SDK::read_driver_name_all(names);

    int index = 0;
    wstring* name = &names[0];
    while (index++ - num_devices)
        wcout << index << ": " << *name++ << endl;
    cout << index << ": [EXIT]" << endl;
	cout << query << endl;

	string s;
	cin >> s;
	long ans = strtol(s.c_str(), nullptr, 0);
	if (ans <= 0 || ans > num_devices + 1) {
		cout << "Did not understand the input." << endl;
		return -1;
	}

	if (ans == index)
		return -2;

    return (int) ans - 1;
}

int main() {
	int device_index = -1; 
	while (device_index == -1)
		device_index = ask_for_device("Which device would you like to test?");
	
	if (device_index == -2)
		return 0;
	
	wcout << "Chosen device: " << ASIO_SDK::read_driver_name(device_index) << endl;

	DAW1 daw;
	ASIODevice soundcard;
	soundcard.attach(device_index, &daw); 

	cout << "Sample rate: " << soundcard.get_sample_rate() << endl;
	cout << "Buffer size: " << soundcard.get_buffer_size() << endl;

	wcout << "Setup complete and currently playing a " << sinewave_hz << " sine tone. Enter something to end the test." << endl;
	cout << endl;
	cout << "Variation in interval between calls to bufferSwitch:" << endl;
	printf("%10s | %10s | %10s", "Min (us)", "Mean (us)", "Max (us)");

	bool running = true;
	thread print_thread(print_process, &running, &daw);

	cout << endl;
	string s;
	cin >> s;
	
	running = false;
	soundcard.detach();
	print_thread.join();

	cout << endl;
	cout << "Enter anything to exit." << endl;
	cin >> s;

	return 0;
}