#pragma once

#include <stdint.h>
#include <vector>
#include <codecvt>
#include <locale>
#include <thread>
#include "asio_sdk/ginclude.h"
#include "Instantiation_and_registry.h"
#include "ASIO_Driver_Interface.h"
#include "ASIO_callback_registry.h"

class AudioHost;

/// <summary>
/// A possibly more comprehensive way of implementing an ASIO device in your program. 
/// This class functions as a lightweight layer over the device, to give easier access to those who are unfamiliar with the ASIO protocol.
/// In terms of performance, this class only impacts the initialization of the device (through attach()). 
/// In terms of streaming-performance, the only step added is a function call to this device's callback functions (namely bufferswitch() etc.).
/// This is necessary for the opportunity to have multiple active ASIO devices.
/// 
/// Note that by default, the device is kept in a RUNNING operational state, as described by the ASIO protocol documentation.
/// Most popular professional audio software contains features - like software instruments - which make this a necessary, comprehensive approach.
/// Equipment will not suffer from this.
/// </summary>
class ASIODevice : private ASIO_callback_registry_entry {
public:

	//ASIODevice(const ASIODevice& a);
	ASIODevice(int deviceIndex, AudioHost* listener, bool default_reset_mechanism);
	ASIODevice(int deviceIndex, AudioHost* listener); // Calls attach()
	ASIODevice(bool default_reset_mechanism);         // Creates an unattached instance
	ASIODevice();                                     // Creates an unattached instance
	~ASIODevice();                                    // Calls detach()

	// Attaches a device to this instance. The device is put in a RUNNING operational state, as described by the ASIO protocol documentation.
	void attach(int deviceIndex, AudioHost* listener);
	void attach(int deviceIndex, AudioHost* listener, ASIOBufferInfo* channelConfig, long numChannels);
	void detach();
	
	void reset();
	void asynch_reset();
	const bool& is_default_reset_mechanism_enabled() const { return default_reset_mechanism; }
	void disable_default_reset_mechanism();
	void  enable_default_reset_mechanism();

	void setChannelStatus(bool isInput, int channel, bool open); // Useful when you want to open just one channel. Briefly stops the audio-stream from and to the device to commit the change.
	void setChannelStatus_noCommit(bool isInput, int channel, bool open); // Opens and closes audio channels, but does not commit the change. Only stores it for the next call to 'setChannelStatus_commit()'.
	void setChannelStatus_commit(); // Briefly stops the audio-stream from the device to commit the changes made through calls to 'setChannelStatus_noCommit()'.
	bool isChannelOpen(bool isInput, int channel) const; // Returns true if the requested channel is currently active. Does not take into account any calls to 'setChannelStatus_noCommit()'.

	void setSampleRate(long sampleRate);
	void setBufferSize(long bufferSize);

	const int32_t* get_input (const int& channel) const { return  inputBuffers[double_buffer_index][channel]; }
	      int32_t* get_output(const int& channel) const { return outputBuffers[double_buffer_index][channel]; }
	
	void silence_output();

	// Getters

	const int&          	 get_device_index()             const { return device_index; }
	const AudioHost* 		 get_listener() 			    const { return listener; }
	const std::wstring& 	 get_name() 					const { return name; }
	const IASIO*        	 get_device_handle()            const { return device_handle; }
	const boolean& 			 is_attached() 					const { return attached; }
	const long& 			 get_sample_rate() 				const { return sampleRate; }
	const long& 			 get_buffer_size() 				const { return bufferSize; }
	const std::vector<long>& get_compatible_sampleRates() 	const { return compatible_sampleRates; }
	const std::vector<long>& get_compatible_bufferSizes() 	const { return compatible_bufferSizes; }
	const long& 			 get_num_inputs() 				const { return numInputs; }
	const long& 			 get_num_outputs() 				const { return numOutputs; }
	const long& 			 get_num_open_inputs() 			const { return numOpenInputs; }
	const long& 			 get_num_open_outputs() 		const { return numOpenOutputs; }
	int32_t** 				 get_input_buffers(int doubleBufferIndex)  const { return  inputBuffers[doubleBufferIndex]; }
	int32_t** 				 get_output_buffers(int doubleBufferIndex) const { return outputBuffers[doubleBufferIndex]; }
	const std::string&       get_error_message()             const { device_handle->getErrorMessage(this->error_message); return error_message; }

	// Methods exposed from the driver interface.

    long      get_driver_version() { return device_handle->getDriverVersion(); }
//  void      get_error_message(char* std::string) { device_handle->getErrorMessage(std::string); }
    ASIOError start() { return device_handle->start(); }
    ASIOError stop() { return device_handle->stop(); }
//  ASIOError getChannels(long* numInputChannels, long* numOutputChannels) { return device_handle->getChannels(numInputChannels, numOutputChannels); }
//  ASIOError getLatencies(long* inputLatency, long* outputLatency) { return device_handle->getLatencies(inputLatency, outputLatency); }
//  ASIOError getBufferSize(long* minSize, long* maxSize, long* preferredSize, long* granularity) { return device_handle->getBufferSize(minSize, maxSize, preferredSize, granularity); }
//  ASIOError canSampleRate(ASIOSampleRate  sampleRate) { return device_handle->canSampleRate(sampleRate); }
//  ASIOError getSampleRate(ASIOSampleRate* sampleRate) { return device_handle->getSampleRate(sampleRate); }
//  ASIOError setSampleRate(ASIOSampleRate  sampleRate) { return device_handle->setSampleRate(sampleRate); }
    ASIOError get_clock_sources(ASIOClockSource* clocks, long* numSources) { return device_handle->getClockSources(clocks, numSources); }
    ASIOError set_clock_sources(long reference) { return device_handle->setClockSource(reference); }
    ASIOError get_sample_position(ASIOSamples* sPos, ASIOTimeStamp* tStamp) { return device_handle->getSamplePosition(sPos, tStamp); }
    ASIOError get_channel_info(ASIOChannelInfo* info) { return device_handle->getChannelInfo(info); }
//  ASIOError createBuffers(ASIOBufferInfo* bufferInfos, long numChannels, long bufferSize, ASIOCallbacks* callbacks) { return device_handle->createBuffers(bufferInfos, numChannels, bufferSize, callbacks); }
//  ASIOError disposeBuffers() { return device_handle->disposeBuffers(); }
    ASIOError control_panel() { return device_handle->controlPanel(); }
    ASIOError future(long selector, void* opt) { return device_handle->future(selector, opt); }
    ASIOError output_ready() { return device_handle->outputReady(); }

protected:
	static const long ALL_SAMPLE_RATES[23];

	AudioHost* listener; // Pointer to an audio processing object. All callbacks from the device are sent to the object pointed to by this pointer. Can be changed at any time.

	int device_index;
	std::wstring name;
	IASIO* device_handle;
	boolean attached; // True if this instance is attached to an ASIO driver and running an audio stream. Only changed by 'attach' and 'detach'.

	long sampleRate;
	long bufferSize;
	std::vector<long> compatible_sampleRates;
	std::vector<long> compatible_bufferSizes;

	long numInputs;
	long numOutputs;
	long numOpenInputs; 
	long numOpenOutputs;

	int32_t** inputBuffers[2]; // Structured like this: [doubleBufferIndex][channel][sample]
	int32_t** outputBuffers[2];// Structured like this: [doubleBufferIndex][channel][sample]
	long double_buffer_index;

	bool reset_device = false;
	bool default_reset_mechanism = false;
	std::thread* reset_thread;
	void reset_thread_process();

	// The current channel-configuration of this device. To learn more, search "ASIOBufferInfo" in the ASIO documentation provided by Steinberg (a copy provided in this repository)
	std::vector<ASIOBufferInfo> currentChnCfg;
	// Represents the channels to be opened/closed in the next call to 'setChannelStatus_commit()'. A channel number of -1 denotes a channel to be closed.
	// The goal of using an ASIOBufferInfo-std::vector instead of a simple bool-array, is to make a cleaner looking code and to slightly boost performance.
	std::vector<ASIOBufferInfo> nextChnCfg;

	char* error_message = new char[124];

	void readNumInOut();
	void readSampleRate();
	void readBufferSize();
	void findCompatibleSampleRates();
	void sortBuffers(ASIOBufferInfo* channelConfig, long numChannels); // Sorts the buffer pairs from each element in 'channelsBufferInfo' into 'inputBuffers' and 'outputBuffers'

#pragma region Stuff only relevant to the callback registry
	int callback_registry_index = -1;
	ASIOCallbacks callbacks;
	void registerCallbacks();
	void deRegisterCallbacks();

#pragma region Inherited methods from ASIO_callback_registry_entry
	void		bufferSwitch(long doubleBufferIndex, ASIOBool directProcess);
	void		sampleRateDidChange(ASIOSampleRate sRate);
	long		asioMessage(long selector, long value, void* message, double* opt);
	ASIOTime*   bufferSwitchTimeInfo(ASIOTime* params, long doubleBufferIndex, ASIOBool directProcess);
#pragma endregion
#pragma endregion

private:
	// This private member of ASIODevice is needed to convert 'ASIODevice::name' from std::wstring to std::string
	static std::wstring_convert<std::codecvt_utf8<wchar_t>, wchar_t> cv;
	static void convert_ASIOSampleRate(ASIOSampleRate sRate, long* target);
};

class AudioHost {
public:
	virtual void		bufferSwitch(long doubleBufferIndex, bool directProcess, ASIODevice* caller) { return; }
	virtual void		sampleRateDidChange(long sampleRate, ASIODevice* caller) { return; }
	virtual long		asioMessage(long selector, long value, void* message, double* opt, ASIODevice* caller) { return 0L; }
	virtual ASIOTime*   bufferSwitchTimeInfo(ASIOTime* params, long doubleBufferIndex, bool directProcess, ASIODevice* caller) { return nullptr; }
	virtual bool        device_reset_init(ASIODevice* caller) { return true; };
	virtual void        device_reset_end(ASIODevice* caller) { return; } ;
};