#pragma once

#include <string>
#include "ASIO_Driver_Interface.h"
#include "ASIO_callback_registry.h"

class AudioHost;

/*This class is an example of the bare minimum you need to be able to use multiple ASIO devices at once.
Here is a list of requirements:
 * Class is implementing the interface 'ASIO_callback_registry_entry'
 * Class registers itself in the callback registry (preferrably in the constructor)
 * Class un-registers itself from the callback registry (preferrably in the destructor)
 * Class stores it's index from the callback registry to be able to un-register.
 * Class stores a reference to an ASIOCallbacks-object, containing pointers to the callbacks recieved from 'ASIO_callback_registry::getCallbackPointers()' (This is the object you send to 'device->createBuffers()' as the last argument)

Also, the class should hold a pointer to the actual device itself. This is not critical, but makes way more sense in this example:
bufferswitch-calls from the device are forwarded to the 'listener' (the DAW or any other audio processing object) with a pointer to 'this'.
The intention is to make critical characteristics - such as sample rate - available to the audio-calculation.
This allows the 'listener' to adapt to multiple devices on-the-go without a massive reduction in performance.

Another solution which might give better performance, is to combine all devices in a common aggregate where they have an equal buffer size and sample rate.
*/
class BasicASIODevice : private ASIO_callback_registry_entry {
public:
	IASIO* device; // The actual asio device. This object is the closest we can get to talking directly with the device.
	ASIOCallbacks callbacks; // Pointers to the callback functions that 'device' is registered to call.

	
	AudioHost* listener; // This variable must be set before 'device->start()' is called! (Done via the 'init' method)


	BasicASIODevice(AudioHost* listener);
	// WARNING: Sets 'listener' to nullptr, resulting in access violation error at runtime if 'device->start()' is called.
	BasicASIODevice();
	~BasicASIODevice();

private:
	int registry_index = -1; // Should only be read from. Gets set through the 'init()' method, and set to -1 by '~BasicASIODevice()'.

	// Returns either the index recieved from 'register_callbacks()', or -1 if there are no more slots avilable in the callback registry.
	int init(AudioHost* listener);
	/* Adds this object as an entry in the callback registry, and stores it's position inside the callback registry in 'index'.
	Also stores pointers to the correct callbacks (indicated to by 'index') in 'callbacks',
	which should be used as the last argument in 'device->createBuffers'.
	*/
	int  register_callbacks();
	void set_listener(AudioHost* listener);

#pragma region Inherited methods from ASIO_callback_registry_entry
	void		bufferSwitch(long doubleBufferIndex, ASIOBool directProcess);
	void		sampleRateDidChange(ASIOSampleRate sRate);
	long		asioMessage(long selector, long value, void* message, double* opt);
	ASIOTime*   bufferSwitchTimeInfo(ASIOTime* params, long doubleBufferIndex, ASIOBool directProcess);
#pragma endregion
};



// Only classes that extend this virtual class are able to process audio of a BasicASIODevice.
class AudioHost {
public:
	virtual void		bufferSwitch(long doubleBufferIndex, ASIOBool directProcess, BasicASIODevice* caller) = 0;
	virtual void		sampleRateDidChange(ASIOSampleRate sRate, BasicASIODevice* caller) = 0;
	virtual long		asioMessage(long selector, long value, void* message, double* opt, BasicASIODevice* caller) = 0;
	virtual ASIOTime*   bufferSwitchTimeInfo(ASIOTime* params, long doubleBufferIndex, ASIOBool directProcess, BasicASIODevice* caller) = 0;
};

// Default 
class AudioHost_default : public AudioHost {
public:
	void		bufferSwitch(long doubleBufferIndex, ASIOBool directProcess, BasicASIODevice* caller) override;
	void		sampleRateDidChange(ASIOSampleRate sRate, BasicASIODevice* caller) override;
	long		asioMessage(long selector, long value, void* message, double* opt, BasicASIODevice* caller) override;
	ASIOTime*   bufferSwitchTimeInfo(ASIOTime* params, long doubleBufferIndex, ASIOBool directProcess, BasicASIODevice* caller) override;
};