
#include <stdexcept>
#include <iostream>
#include <future>
#include "ASIO_device.h"

using namespace std;

const long ASIODevice::ALL_SAMPLE_RATES[23] = {
	8000,
	11025,
	16000,
	22050,
	32000,
	37800,
	44056,
	44100,
	47250,
	48000,
	50000,
	50400,
	64000,
	88200,
	96000,
	176400,
	192000,
	352800,
	384000,
	2822400,
	5644800,
	11289600,
	22579200
};
wstring_convert<codecvt_utf8<wchar_t>, wchar_t> ASIODevice::cv;

ASIODevice::ASIODevice(int deviceIndex, AudioHost* listener) : ASIODevice(deviceIndex, listener, true) {
	
}

ASIODevice::ASIODevice(int deviceIndex, AudioHost* listener, bool default_reset_mechanism) : ASIODevice(default_reset_mechanism) {
	if (default_reset_mechanism)
		enable_default_reset_mechanism();
	attach(deviceIndex, listener);
}

ASIODevice::ASIODevice() : ASIODevice(true) {
	
}

ASIODevice::ASIODevice(bool default_reset_mechanism) {
	device_handle = nullptr;
	name = L"";
	attached = false;

	sampleRate = -1;
	bufferSize = -1;

	numInputs = -1;
	numOutputs = -1;

	if (default_reset_mechanism)
		enable_default_reset_mechanism();
}

ASIODevice::~ASIODevice() {
	disable_default_reset_mechanism();
	detach();
}

void ASIODevice::attach(int deviceIndex, AudioHost* listener) {
	ASIOBufferInfo initialBuffers[3];
	// Inputs
	initialBuffers[0].channelNum = 0;
	initialBuffers[0].isInput = true;
	// Outputs
	initialBuffers[1].channelNum = 0;
	initialBuffers[1].isInput = false;
	initialBuffers[2].channelNum = 1;
	initialBuffers[2].isInput = false;

	attach(deviceIndex, listener, initialBuffers, 3);
}

void ASIODevice::attach(int deviceIndex, AudioHost* listener, ASIOBufferInfo* channelConfig, long numChannels) {
	if (attached) {
		throw exception("This instance is already attached to a device!");
		return;
	}
		
	this->device_index = deviceIndex;
	this->listener = listener;
	this->name = ASIO_SDK::read_driver_name(deviceIndex); // Reading the device's name from the Windows registry.
	this->registerCallbacks();

	// Create an instance of the device driver (using Windows COM-library) from 
	// the driver's CLSID (extracted from Windows registry) and store it in 'deviceHandle'
	HRESULT err = ASIO_SDK::createInstance(ASIO_SDK::read_driver_CLSID(deviceIndex), &device_handle);

	device_handle->AddRef(); // Update the internal counter of the driver.
	DWORD processID = GetCurrentProcessId();
	ASIOError aerr = device_handle->init((void*)&processID); // Give the driver an ID handle to this calling process.

	readNumInOut();
	readSampleRate();
	readBufferSize();
	findCompatibleSampleRates();

	inputBuffers[0]  = new int32_t * [numInputs];
	inputBuffers[1]  = new int32_t * [numInputs];
	outputBuffers[0] = new int32_t * [numOutputs];
	outputBuffers[1] = new int32_t * [numOutputs];
	// Reset all channel buffers
	for (int i = 0; i < numInputs; i++) {
		inputBuffers[0][i] = nullptr;
		inputBuffers[1][i] = nullptr;
	} for (int i = 0; i < numOutputs; i++) {
		outputBuffers[0][i] = nullptr;
		outputBuffers[1][i] = nullptr;
	}

	// Initialize the buffer that conatins buffer-info
	currentChnCfg.reserve((size_t)numInputs + numOutputs);
	ASIOBufferInfo input  = { true, -1 };
	ASIOBufferInfo output = { false, -1 };
	for (int i = 0; i < numInputs; i++)
		currentChnCfg.push_back(input);
	for (int i = 0; i < numOutputs; i++)
		currentChnCfg.push_back(output);

	// Add arguments to the buffer
	for (int i = 0; i < numChannels; i++) {
		int index = channelConfig[i].channelNum + (channelConfig[i].isInput ? 0 : numInputs);
		currentChnCfg[index] = channelConfig[i];
	}
	nextChnCfg = currentChnCfg;

	device_handle->createBuffers(channelConfig, numChannels, bufferSize, &callbacks); // Create buffers for every channel.
	sortBuffers(channelConfig, numChannels); // Gather all channel buffers together in two arrays (one for input, one for output). Each array is structured like this: [doubleBufferIndex][channel][data]
	silence_output();
	
	device_handle->start(); // Start the audio streaming
	attached = true;
}

void ASIODevice::detach() {
	if (!attached)
		return;
	attached = false;

	device_handle->stop();
	device_handle->disposeBuffers();
	device_handle->Release();
	deRegisterCallbacks();
	device_handle->Release();

	// Reset values
	name.clear();
	bufferSize = 0;
	sampleRate = 0;
	numInputs  = 0;
	numOutputs = 0;
	numOpenInputs  = 0;
	numOpenOutputs = 0;
	compatible_sampleRates.clear();
	compatible_bufferSizes.clear();
	nextChnCfg.clear();

	// Remove references
	device_handle = nullptr;
	listener = nullptr;
	callbacks.bufferSwitch = nullptr;
	callbacks.sampleRateDidChange = nullptr;
	callbacks.asioMessage = nullptr;
	callbacks.bufferSwitchTimeInfo = nullptr;

	// Avoid memory leaks
	delete( inputBuffers[0]);
	delete( inputBuffers[1]);
	delete(outputBuffers[0]);
	delete(outputBuffers[1]);
}

void ASIODevice::reset() {
	AudioHost* host = this->listener;
	int devIndex = this->device_index;


	if (listener->device_reset_init(this) == false)
		return;
	this->detach();
	this->attach(devIndex, host);
	this->readBufferSize();
	listener->device_reset_end(this);
}

void ASIODevice::asynch_reset() {
	if (!default_reset_mechanism)
		throw exception("The default reset mechanism is not enabled.");
	reset_device = true;
}

void ASIODevice::disable_default_reset_mechanism() {
	if (!default_reset_mechanism)
		return;
	default_reset_mechanism = false;
	reset_thread->join();
	delete(reset_thread);
}

void ASIODevice::enable_default_reset_mechanism() {
	if (default_reset_mechanism)
		return;
	default_reset_mechanism = true;
	reset_thread = new thread(&ASIODevice::reset_thread_process, this);
}

void ASIODevice::reset_thread_process() {
	while (default_reset_mechanism) {
		if (reset_device) {
			reset_device = false;
			reset();
		}
		else Sleep(1000);
	}
}


void ASIODevice::readNumInOut() {
	device_handle->getChannels(&numInputs, &numOutputs);
}

void ASIODevice::readSampleRate() {
	ASIOSampleRate sr;
	device_handle->getSampleRate(&sr);
	sampleRate = sr;
}

void ASIODevice::readBufferSize() {
	long min, max, gran;
	device_handle->getBufferSize(&min, &max, &bufferSize, &gran);

	int arrSize = log2(max) - log2(min) + 1;
	compatible_bufferSizes.clear();
	compatible_bufferSizes.reserve(arrSize);

	for (int i = min; i <= max; i *= 2)
		compatible_bufferSizes.push_back(i);
	compatible_bufferSizes.shrink_to_fit();
}

void ASIODevice::findCompatibleSampleRates() {
	compatible_sampleRates.clear();
	int arraySize = sizeof(ALL_SAMPLE_RATES) / sizeof(ALL_SAMPLE_RATES[0]);

	compatible_sampleRates.reserve(arraySize);
	for (int i = 0; i < arraySize; i++)
		if (device_handle->canSampleRate(ALL_SAMPLE_RATES[i]) == ASE_OK)
			compatible_sampleRates.push_back(ALL_SAMPLE_RATES[i]);
	compatible_sampleRates.shrink_to_fit();
}



void ASIODevice::sortBuffers(ASIOBufferInfo* channelConfig, long numChannels) {
	for (long i = 0; i < numChannels; i++) {
		long channel = channelConfig[i].channelNum;
		if (channelConfig[i].isInput) {
			inputBuffers[0][channel] = reinterpret_cast<int32_t*>(channelConfig[i].buffers[0]);
			inputBuffers[1][channel] = reinterpret_cast<int32_t*>(channelConfig[i].buffers[1]);
			numOpenInputs = numOpenInputs > channel ? numOpenInputs : channel;
		}
		else {
			outputBuffers[0][channel] = reinterpret_cast<int32_t*>(channelConfig[i].buffers[0]);
			outputBuffers[1][channel] = reinterpret_cast<int32_t*>(channelConfig[i].buffers[1]);
			numOpenOutputs = numOpenOutputs > channel ? numOpenOutputs : channel;
		}
	}
}

void ASIODevice::silence_output() {
	if (!outputBuffers[0])
		return;

	for (int ch = 0; ch < numOutputs; ch++) {
		if (outputBuffers[0][ch]) {
			memset(outputBuffers[0][ch], 0, bufferSize * sizeof(int32_t));
			memset(outputBuffers[1][ch], 0, bufferSize * sizeof(int32_t));
		}
	}
}



// Elements relating to the ASIO_callback_registry -solution.
void ASIODevice::registerCallbacks() {
	this->callback_registry_index = ASIO_callback_registry::registerEntry(this);
	ASIO_callback_registry::getCallbackPointers(callback_registry_index, &callbacks);
}

void ASIODevice::deRegisterCallbacks() {
	ASIO_callback_registry::removeEntry(callback_registry_index);
	callback_registry_index = -1;
}



void ASIODevice::bufferSwitch(long doubleBufferIndex, ASIOBool directProcess) {
	this->double_buffer_index = doubleBufferIndex;
	listener->bufferSwitch(doubleBufferIndex, directProcess, this);
}

void ASIODevice::sampleRateDidChange(ASIOSampleRate sRate) {
	ASIODevice::convert_ASIOSampleRate(sRate, &this->sampleRate);
	listener->sampleRateDidChange(this->sampleRate, this);
}

long ASIODevice::asioMessage(long selector, long value, void* message, double* opt) {
	char* message_char = (char*)message;

	// Device asking for supported selectors
	if (selector == kAsioSelectorSupported)
		switch (value) {
		case kAsioResetRequest:
			if (this->default_reset_mechanism)
				return 1L;
			else break;
		//case kAsioBufferSizeChange: return 1L;
		//case kAsioEngineVersion:    return 1L;
		//case kAsioSupportsTimeInfo: return 1L;
		//case kAsioResyncRequest:    return 0L;
		//case kAsioLatenciesChanged: return 0L;
		//case kAsioOverload:         return 0L;

		default: break;
		}
	else {
		switch (selector) {
		case kAsioBufferSizeChange: this->readBufferSize();
		case kAsioResetRequest:
			if (this->default_reset_mechanism) {
				this->reset_device = true;
				return 1L;
			}
		//case kAsioEngineVersion:    return 2L;
		//case kAsioResyncRequest:    return 0L;
		//case kAsioSupportsTimeInfo: return 0L;
		//case kAsioOverload:         return 0L;
		}
	}
	
	return listener->asioMessage(selector, value, message, opt, this);
}

ASIOTime* ASIODevice::bufferSwitchTimeInfo(ASIOTime* params, long doubleBufferIndex, ASIOBool directProcess) {
	this->double_buffer_index = doubleBufferIndex;
	return listener->bufferSwitchTimeInfo(params, doubleBufferIndex, directProcess, this);
}



void ASIODevice::setSampleRate(long newSampleRate) {
	if (!attached) {
		throw exception("Error setting sample rate: This instance is not attached to any ASIO device.");
		return;
	}

	ASIOError ase = device_handle->canSampleRate(newSampleRate); // Check wether this device supports this sample rate

	switch (ase) {
	case ASE_OK:
		this->sampleRate = newSampleRate;
		device_handle->setSampleRate(newSampleRate);
		return;
	case ASE_NoClock:
		throw invalid_argument("ASIO device '" + cv.to_bytes(name) + "' does not support the sample rate: " + to_string(newSampleRate));
		return;
	case ASE_NotPresent:
		throw exception("Error setting sample rate: The device returned ASE_NotPresent - \"no input/output is present\".");
		return;
	}
}

void ASIODevice::setBufferSize(long newBufferSize) {
	bool invalidBufferSize = true;

	for (int i = 0; i < compatible_bufferSizes.size(); i++)
		if (compatible_bufferSizes[i] == newBufferSize) {
			invalidBufferSize = false;
			break;
		}

	if (invalidBufferSize) {
		throw invalid_argument("Buffersize " + to_string(newBufferSize) + " is not supported by device " + cv.to_bytes(name));
		return;
	}

	this->bufferSize = newBufferSize;
	if (!attached)
		return;
	
	vector<ASIOBufferInfo> channelConfig = currentChnCfg;
	channelConfig.erase(remove_if(channelConfig.begin(), channelConfig.end(), [](ASIOBufferInfo x) {return x.channelNum < 0; }), channelConfig.end());
	channelConfig.shrink_to_fit();

	device_handle->stop();
	device_handle->disposeBuffers();
	device_handle->createBuffers(&channelConfig[0], channelConfig.size(), bufferSize, &callbacks);
	device_handle->start();
}

void ASIODevice::setChannelStatus(bool input, int channel, bool open) {
	setChannelStatus_noCommit(input, channel, open);
	setChannelStatus_commit();
}

void ASIODevice::setChannelStatus_noCommit(bool input, int channel, bool open) {
	if (channel < 0 || channel >= (input ? numInputs : numOutputs)) { // If out of bounds, throw exception
		string channelType = (input ? "Input" : "Output");
		string error_string = channelType + " channel " + to_string(channel) + " does not exist on this device: " + cv.to_bytes(name);
		cerr << error_string;

		throw out_of_range(error_string);
		return;
	}

	int index = channel + (input ? 0 : numInputs);
	nextChnCfg[index].channelNum = open ? channel : -1;
}

void ASIODevice::setChannelStatus_commit() {
	vector<ASIOBufferInfo> channelConfig = nextChnCfg;
	channelConfig.erase(remove_if(channelConfig.begin(), channelConfig.end(), [](ASIOBufferInfo x) {return x.channelNum < 0; }), channelConfig.end());
	channelConfig.shrink_to_fit();

	if (channelConfig.size() < 1) {                                   // At least on channel has to be open at all times for us to be able to run the device.
		throw invalid_argument("Exception while changing channel status: At least on channel has to be open at all times.");
		return;
	}

	// Now we have what we need to pull off the manouver

	device_handle->stop(); // First, pull the device down to the operational state, that allows us to change the channel config.
	device_handle->disposeBuffers();

	// Reset all channel buffers
	for (int i = 0; i < numInputs; i++) {
		inputBuffers[0][i] = nullptr;
		inputBuffers[1][i] = nullptr;
	} for (int i = 0; i < numOutputs; i++) {
		outputBuffers[0][i] = nullptr;
		outputBuffers[1][i] = nullptr;
	}

	ASIOError ase = device_handle->createBuffers(&channelConfig[0], channelConfig.size(), bufferSize, &callbacks);
	switch (ase) {
	case ASE_OK: break;
	case ASE_InvalidMode:
		throw invalid_argument("Exception while changing channel status: Either the requested buffer size is not supported, or there was something wrong with the channel config.");
		return;
	case ASE_NotPresent:
		throw invalid_argument("One of the requested channels are not supported on this device.");
		return;
	}

	sortBuffers(&channelConfig[0], channelConfig.size());
	device_handle->start();

	currentChnCfg = channelConfig; // Copy the new channel config in as the new channel config
}

bool ASIODevice::isChannelOpen(bool input, int channel) const {
	if (channel < 0 || channel >= (input ? numInputs : numOutputs)) { // If out of bounds, throw exception
		wstring channelType = (input ? L"Input" : L"Output");
		wstring error_string = channelType + L" channel " + to_wstring(channel) + L" does not exist on this device: " + name;
		wcerr << error_string;

		throw out_of_range(cv.to_bytes(error_string));
		return false;
	}

	int32_t* inQuestion = (input ? inputBuffers[0][channel] : outputBuffers[0][channel]);

	if (inQuestion != nullptr)
		return true;
	return false;
}

void ASIODevice::convert_ASIOSampleRate(ASIOSampleRate sRate, long* target) {
#if IEEE754_64FLOAT
	*target = (long) sRate;
#else
	memcpy(target, sRate.ieee, 8);
#endif
}