
#include <iostream>
#include <limits>
#include <exception>
#include "SampleJumper.h"

SampleJumper::SampleJumper(const int16_t& hardware_buffersize) {
    this->buffersize = hardware_buffersize * 5;
    this->buffersize_in_bytes = buffersize * SAMPLE_SIZE;
    this->hardware_buffersize = hardware_buffersize;
    this->hardware_buffersize_in_bytes = hardware_buffersize * SAMPLE_SIZE;

    this->buffer = new int32_t[this->buffersize] {0};
    this->temp_buffer = new int32_t[this->buffersize] {0};

    this-> read_cursor = 0;
    this->write_cursor = 0;
    this->roundtrip = false;
    this->ignore_xruns = false;
    this->threshold_min = hardware_buffersize * 2;
    this->threshold_max = hardware_buffersize * 4;
    
    this->samples_offset = 0;
    this->set_max_sample_jump(1 << 16);
}

SampleJumper::~SampleJumper() {
    delete(this->buffer);
    delete(this->temp_buffer);
}


void SampleJumper::writeNext(int16_t samples_jumped, const int32_t* &input,  const uint16_t& samplecount) {
    //if (samples_jumped)
        //std::cout << samples_jumped << " samples jumped." << std::endl;

    // Perform the shortening algorithm if necessary

    //samples_offset += samples_jumped;
    //if (samples_offset < 0) {
    //    if (get_cursor_distance() > abs(samples_offset)) {
    //        int16_t num_jumped = try_compress(abs(samples_offset));
    //        samples_offset -= num_jumped;
    //        if (samples_offset <= 0)
    //            std::cout << "Fixed. W_C= " << write_cursor << std::endl;
    //    } else {
    //        samples_offset = 0;
    //        std::cout << "Too little available data. Ignored." << std::endl;
    //    }
    //    
    //}
        
    // If we are still missing data to send onwards, perform underrun.
    // Check that we have enough space to write.
    uint16_t curDist = get_cursor_distance();
    if (curDist + samplecount > buffersize) {
        this->faster_consumer = false; // Seems like the consumer's clock is faster than the producer's.
        std::cout << "Overrun. Cursor distance: " << curDist << " < " << samplecount << std::endl;
        return;
    }

    // Write the received data to the buffer
    int16_t dist_to_end = this->buffersize - this->write_cursor;
    if (samplecount > dist_to_end) { // Do we have to do the write while closing the loop?
        memcpy(&buffer[write_cursor], input, dist_to_end * SAMPLE_SIZE); // Yes. Write to the end of the ring buffer.
        write_cursor_inc(samplecount); // Set the write_cursor at the end of the data.
        memcpy(&buffer[0], &input[dist_to_end], write_cursor * SAMPLE_SIZE); // Use the absolute value of the write_cursor to denote how many samples we have left to write, beginning at the start of the ring buffer.
    }
    else {
        memcpy(&this->buffer[this->write_cursor], input, samplecount * SAMPLE_SIZE); // Copy all data into the buffer at the position of the write_cursor. 
        write_cursor_inc(samplecount); // Move the write cursor forward.
    }
    // Writing complete
}

void SampleJumper::readNext (int16_t samples_jumped,       int32_t* &output, const uint16_t& samplecount) {
    //if (samples_jumped)
        //std::cout << samples_jumped << " samples jumped." << std::endl;
    
    //Perform the lengthening algorithm if necessary
    
    int samples_to_fix = threshold_min - get_cursor_distance();
    if (samples_to_fix > 0)
        int16_t num_jumped = try_lengthen(samples_to_fix);
        
    // If we are still missing data to send onwards, perform underrun.
    // Check that we have enough data to read.
    uint16_t curDist = get_cursor_distance();
    if (curDist < samplecount) {
        this->faster_consumer = true; // Seems like the consumer's clock is faster than the producer's.
        std::cout << "Underrun. Cursor distance: " << curDist << " < " << samplecount << std::endl;
        return;
    }

    int16_t dist_to_end = this->buffersize - this->read_cursor;
    
    if (samplecount > dist_to_end) {
        memcpy(output, &buffer[read_cursor], dist_to_end * SAMPLE_SIZE);
        read_cursor_inc(samplecount);
        memcpy(&output[dist_to_end], &buffer[0], read_cursor * SAMPLE_SIZE);
    }
    else {
        memcpy(output, &this->buffer[this->read_cursor], samplecount * SAMPLE_SIZE);
        read_cursor_inc(samplecount);
    }
}

void SampleJumper::read_cursor_inc(const uint16_t& inc) {
    uint16_t distance = get_cursor_distance();
    if (distance - inc < 0 && !ignore_xruns)
        throw std::invalid_argument("Buffer underrun! Tried reading from the buffer when the buffer is empty.");

    uint16_t newpos = read_cursor + inc;
    if (newpos < buffersize)
        read_cursor = newpos;
    else
        read_cursor = newpos - buffersize;

    roundtrip = false;
}

void SampleJumper::write_cursor_inc(const uint16_t& inc) {
    if (get_cursor_distance() + inc > buffersize && !ignore_xruns)
        throw std::invalid_argument("Buffer overrun! Tried writing to the buffer when there is no more space in the buffer.");

    uint16_t newpos = write_cursor + inc;
    if (newpos < buffersize)
        write_cursor = newpos;
    else
        write_cursor = newpos - buffersize;
}

void SampleJumper::read_cursor_inc() {
    if (read_cursor == write_cursor && !roundtrip && !ignore_xruns)
        throw std::invalid_argument("Buffer underrun! Tried reading from the buffer when the buffer is empty.");
    
    read_cursor++;
    if (read_cursor >= buffersize)
        read_cursor = 0;

    //read_cursor < buffersize ?  read_cursor++ :  read_cursor = 0;
    roundtrip = false;
}

void SampleJumper::write_cursor_inc() {
    if (write_cursor == read_cursor && roundtrip && !ignore_xruns)
        throw std::invalid_argument("Buffer overrun! Tried writing to the buffer when there is no more space in the buffer.");
    
    write_cursor++;
    if (write_cursor >= buffersize)
        write_cursor = 0;
    
    if (write_cursor == read_cursor) // If the write_cursor is equal to the read_cursor after incrementing its position by one, we have a roundtrip!
        roundtrip = true;
}

inline void SampleJumper::set_threshold_volume(const float  decibels_0DBFS) {
    if (decibels_0DBFS > 0)
        throw std::invalid_argument("The decibel value must be less than or equal to zero.");
    
    this->set_max_sample_jump((int32_t) (sample_0dBFS * pow(10, (decibels_0DBFS/20))));
}

inline void SampleJumper::set_max_sample_jump(const int32_t  sample_val_threshold) {
    this->sample_diff_threshold = sample_val_threshold;
}

uint16_t SampleJumper::try_compress(const uint16_t& num_samples) {
    uint16_t samples_jumped = 0;
    const uint16_t original_read_cursor = read_cursor;
    const uint16_t end = write_cursor; // Mark the end.
    write_cursor = read_cursor; // Align the cursors.
    ignore_xruns = true;

    while (read_cursor != end) {
        read_cursor_inc();
        if (abs(buffer[write_cursor] - buffer[read_cursor]) < sample_diff_threshold && num_samples - samples_jumped > 0) {
            read_cursor_inc(); // Skip one sample
            samples_jumped++;
        }
        
        write_cursor_inc();
        buffer[write_cursor] = buffer[read_cursor];
    }

    read_cursor = original_read_cursor; // Reset the reader head back to original position.
    ignore_xruns = false;
    return samples_jumped;
}

void printBuffer(int32_t* buffer, uint16_t bufferSize) {
    std::cout << "[" << buffer[0] << ", ";

    for (int i = 1; i < bufferSize - 1; i++)
        std::cout << buffer[i] << ", ";
    
    std::cout << buffer[bufferSize - 1] << "]" << std::endl;
}

uint16_t SampleJumper::try_lengthen(const uint16_t& num_samples) {
    uint16_t samples_jumped = 0;
    const uint16_t original_read_cursor = read_cursor;
    const uint16_t original_write_cursor = write_cursor;
    const uint16_t iteration_length = get_cursor_distance();
    write_cursor = read_cursor;
    ignore_xruns = false;

    // memcpy(temp_buffer, buffer, buffersize_in_bytes); // Copy all data to the temporary buffer
    
    for (int i = 0; i < iteration_length; i++) {
        temp_buffer[write_cursor] = buffer[read_cursor]; // First, write the current sample to output.
        write_cursor_inc(); // Place the writer head at the next sample.
        
        // Do we need to alter the buffer?
        if (num_samples - samples_jumped > 0) {
            // Yes!
            // If the difference between the current sample and the next sample is small enough,
            // insert a duplicate of the current sample!
            int32_t sample_diff = abs(buffer[write_cursor] - buffer[read_cursor]);
            if (sample_diff < sample_diff_threshold) {
                temp_buffer[write_cursor] = buffer[read_cursor]; // Insert a duplicate of the current sample
                samples_jumped++; // Record the change
                write_cursor_inc(); // Place the writer head at the next sample
            }
        }

        read_cursor_inc(); // Place the reader head at the next sample.
    }

    memcpy(buffer, temp_buffer, buffersize_in_bytes); // Overwrite buffer with our new data.
    read_cursor  = original_read_cursor;  // Reset the reader head back to original position.
    
    const int write_cursor_curr_pos = write_cursor;
    write_cursor = original_read_cursor;
    write_cursor_inc(iteration_length + samples_jumped);
    const int expected_pos = write_cursor;
    write_cursor = write_cursor_curr_pos;
    if (write_cursor != expected_pos)
        throw std::exception("The write cursor should not have moved after this function");

    ignore_xruns = false;
    return samples_jumped;
}

inline uint16_t SampleJumper::get_cursor_distance() const {
    uint16_t result = roundtrip ? buffersize : 0;

    if (read_cursor <= write_cursor)
        result += write_cursor - read_cursor;
    else
        result += buffersize - (read_cursor - write_cursor);

    return result;
}

uint16_t SampleJumper::MOVE() {
    uint16_t num_samples = 64;
    uint16_t samples_jumped = 0;
    const uint16_t original_read_cursor = read_cursor;
    const uint16_t original_write_cursor = write_cursor;
    const uint16_t iteration_length = get_cursor_distance();
    write_cursor = read_cursor;
    ignore_xruns = false;

    // memcpy(temp_buffer, buffer, buffersize_in_bytes); // Copy all data to the temporary buffer
    
    for (int i = 0; i < iteration_length; i++) {
        temp_buffer[write_cursor] = buffer[read_cursor]; // First, write the current sample to output.
        write_cursor_inc(); // Place the writer head at the next sample.
        
        // Do we need to alter the buffer?
        if (num_samples - samples_jumped > 0) {
            // Yes!
            // If the difference between the current sample and the next sample is small enough,
            // insert a duplicate of the current sample!
            int32_t sample_diff = abs(buffer[write_cursor] - buffer[read_cursor]);
            if (sample_diff < sample_diff_threshold) {
                //temp_buffer[write_cursor] = buffer[read_cursor]; // Insert a duplicate of the current sample
                samples_jumped++; // Record the change
                //write_cursor_inc(); // Place the writer head at the next sample
            }
        }

        read_cursor_inc(); // Place the reader head at the next sample.
    }

    memcpy(buffer, temp_buffer, buffersize_in_bytes); // Overwrite buffer with our new data.
    read_cursor  = original_read_cursor;  // Reset the reader head back to original position.

    if (write_cursor != original_write_cursor)
        throw std::exception("The write cursor should not have moved after this function");

    ignore_xruns = false;
    return samples_jumped;
}