#include "exchange/RingBuffer.h"

#define RB_DATA int32_t
#define INDEX int

RingBuffer::RingBuffer(const INDEX& buffer_size) : buffer_size(buffer_size), INDEX_BYTE_SIZE(sizeof(INDEX)) {
    buffer = new RB_DATA[buffer_size];
    
    _ignore_xruns = false;
    read_cursor = 0;
    write_cursor = 0;
    cursor_distance = 0;
    roundtrip = false;
}

RingBuffer::~RingBuffer() {
    delete(buffer);
}

void RingBuffer::read_next(RB_DATA* destination, const INDEX& amount) {
    lock.lock();

    if (!_ignore_xruns)
        if (cursor_distance - amount < 0)
            throw std::underflow_error("Buffer underflow! Tried reading more elements than the amount of available data.");

    INDEX dist_to_end = buffer_size - read_cursor;
    
    if (amount > dist_to_end) {
        memcpy(destination, &buffer[read_cursor], dist_to_end * INDEX_BYTE_SIZE);
        read_cursor_inc(amount);
        memcpy(&destination[dist_to_end], &buffer[0], read_cursor * INDEX_BYTE_SIZE);
    } else {
        memcpy(destination, &buffer[read_cursor], amount * INDEX_BYTE_SIZE);
        read_cursor_inc(amount);
    }

    lock.unlock();
}

void RingBuffer::write_next(const RB_DATA* source, const INDEX& amount) {
    lock.lock();

    if (!_ignore_xruns)
        if (cursor_distance + amount > buffer_size)
            throw std::overflow_error("Buffer overflow! Not enough available buffer space.");
    
    INDEX dist_to_end = buffer_size - write_cursor;
    
    if (amount > dist_to_end) {
        memcpy(&buffer[write_cursor], source, dist_to_end * INDEX_BYTE_SIZE);
        write_cursor_inc(amount);
        memcpy(&buffer[0], &source[dist_to_end], write_cursor * INDEX_BYTE_SIZE);
    } else {
        memcpy(&buffer[write_cursor], source, amount * INDEX_BYTE_SIZE);
        write_cursor_inc(amount);
    }

    lock.unlock();
}

void RingBuffer::ignore_xruns(const bool& ignore_xruns) {
    lock.lock();
    this->_ignore_xruns = ignore_xruns;
    lock.unlock();
}

const bool& RingBuffer::ignoring_xruns() const {
    return _ignore_xruns;
}

void RingBuffer::read_cursor_inc(const INDEX& increment) {
    if (increment < 0)
        throw std::invalid_argument("Increment must be a positive value.");
    
    cursor_distance -= increment;

    if (!_ignore_xruns)
        if (cursor_distance < 0) {
            cursor_distance += increment;
            throw std::underflow_error("Buffer underflow! Tried reading more elements than the amount of available data.");
        }
    
    read_cursor += increment;
    if (read_cursor >= buffer_size)
        read_cursor -= buffer_size;

    if (roundtrip)
        roundtrip = false;
}

void RingBuffer::write_cursor_inc(const INDEX& increment) {
    if (increment < 0)
        throw std::invalid_argument("Increment must be a positive value.");
    
    cursor_distance += increment;
    
    if (!_ignore_xruns)
        if (cursor_distance > buffer_size) {
            cursor_distance -= increment;
            throw std::overflow_error("Buffer overflow! Not enough available buffer space.");
        }
    
    write_cursor += increment;
    if (write_cursor >= buffer_size)
        write_cursor -= buffer_size;
    
    if (write_cursor == read_cursor) // If the write_cursor is equal to the read_cursor after incrementing its position by one, we have a roundtrip!
        roundtrip = true;
}

void RingBuffer::read_cursor_dec(const INDEX& decrement) {
    if (decrement < 0)
        throw std::invalid_argument("Decrement must be a positive value.");
    
    cursor_distance += decrement;

    if (!_ignore_xruns)
        if (cursor_distance > buffer_size) {
            cursor_distance -= decrement;
            throw std::overflow_error("Buffer overflow! Not enough available buffer space.");
        }
    
    read_cursor -= decrement;
    if (read_cursor < 0)
        read_cursor += buffer_size;

    if (write_cursor == read_cursor) // If the write_cursor is equal to the read_cursor after incrementing its position by one, we have a roundtrip!
        roundtrip = true;
}

void RingBuffer::write_cursor_dec(const INDEX& decrement) {
    if (decrement < 0)
        throw std::invalid_argument("Decrement must be a positive value.");
    
    cursor_distance -= decrement;
    
    if (!_ignore_xruns)
        if (cursor_distance < 0) {
            cursor_distance += decrement;
            throw std::underflow_error("Buffer underflow! Not enough available data.");
        }
    
    write_cursor -= decrement;
    if (write_cursor < 0)
        write_cursor += buffer_size;
    
    if (roundtrip)
        roundtrip = false;
}

INDEX RingBuffer::get_data(RB_DATA* destination) const {
    if (read_cursor == write_cursor && !roundtrip)
        return 0;

    INDEX dist_to_end = buffer_size - read_cursor;

    if (read_cursor < write_cursor) {
        memcpy(destination, &buffer[read_cursor], cursor_distance * INDEX_BYTE_SIZE);
    } else {
        memcpy(destination, &buffer[read_cursor], dist_to_end * INDEX_BYTE_SIZE);
        memcpy(&destination[dist_to_end], &buffer[0], write_cursor * INDEX_BYTE_SIZE);
    }

    return cursor_distance;
}

void RingBuffer::set_data(const RB_DATA* source, INDEX size) {
    if (size > buffer_size)
        throw std::invalid_argument("Size-parameter cannot be larger than the internal buffer size.");
    
    INDEX dist_to_end = buffer_size - read_cursor;

    if (dist_to_end >= size) {
        memcpy(&buffer[read_cursor], source, size * INDEX_BYTE_SIZE);
        write_cursor = read_cursor;
        cursor_distance = 0;
        write_cursor_inc(size);
    }
    else {
        memcpy(&buffer[read_cursor], source, dist_to_end * INDEX_BYTE_SIZE);
        memcpy(&buffer[0], &source[dist_to_end], (size - dist_to_end) * INDEX_BYTE_SIZE);
    }


}

RB_DATA* RingBuffer::get_buffer() const {
    return buffer;
}

const INDEX& RingBuffer::get_data_amount() const {
    return cursor_distance;
}

const INDEX& RingBuffer::get_read_cursor() const {
    return read_cursor;
}

const INDEX& RingBuffer::get_write_cursor() const {
    return write_cursor;
}

#undef RB_DATA
#undef INDEX